<?php

/**
 * SPDX-License-Identifier: EUPL-1.2
 * Authors: see /README.md
 */

namespace Seacms\ComposerInstaller;

use Composer\Composer;
use Composer\Config;
use Composer\Json\JsonFile;
use Composer\IO\IOInterface;
use Composer\Repository\PathRepository;
use Composer\Repository\RepositoryManager;
use Composer\Repository\VcsRepository;
use Exception;
use Composer\Script\Event as ScriptEvent;
use Composer\Script\ScriptEvents;
use Seacms\ComposerInstaller\Handler;
use Seacms\ComposerInstaller\NotFoundFileException;
use Seacms\ComposerInstaller\ParentPlugin;

/**
 * Composer plugin to install or not packages from local path
 */

class Plugin extends ParentPlugin
{
    /**
     * composer instance
     * @var Composer
     */
    protected $composer;
    /**
     * io instance
     * @var IOInterface
     */
    protected $io;
    protected $handler;

    /**
     * Apply plugin modifications to Composer
     *
     * @param Composer $composer
     * @param IOInterface $io
     */
    public function activate(Composer $composer, IOInterface $io) {
        $this->composer = $composer;
        $this->io = $io;
        if (!self::IS_TOP_FLOOR_PLUGIN){
            // do nothing
            $this->handler = null;
            return ;
        }
        $this->handler = new Handler($composer, $io);
    }

    /**
     * manage preupdate event
     * @param ScriptEvent $event
     */
    public function onPreUpdate(ScriptEvent $event)
    {
        if (!is_null($this->handler)){
            try {
                list('config' => $config,'path' => $path) = $this->getConfigJsonFile();
                $this->updateRepositoryManager($config);
                // $composer->getInstallationManager()->addInstaller(new LocalInstaller($io,$composer));
            } catch (NotFoundFileException $ex){
                // do nothing
            }
        }
    }
    
    /**
     * get config.json File from root path
     * @return array ['config' => $config,'path'=>path]
     * @throws NotFoundFileException
     */
    protected function getConfigJsonFile(): array
    {
        $file = new JsonFile('./main-config.json');
        if (!$file->exists()){
            throw new NotFoundFileException("main-config.json file not existing");
        }
        $config = $file->read();
        if (!is_array($config)){
            $this->io->write('<error>Error loading main-config.json : it should contain an array</error>');
            throw new NotFoundFileException("main-config.json file not existing");
        }
        return ['config'=>$config,'path'=>$file->getPath()];
    }
    
    /**
     * update updateRepositoryManager with configFile
     * @param array $config
     */
    protected function updateRepositoryManager(array $config)
    {
        $correspondances = $this->extractAvailableCorrespondances($config);
        if (!empty($correspondances)){
            $repositoryManager = $this->composer->getRepositoryManager();
            $repos = $repositoryManager->getRepositories();
            foreach ($repos as $repo) {
                if ($repo instanceof VcsRepository){
                    try {
                        $this->prependRepoIfFound($repositoryManager,$correspondances,$repo);
                    } catch (Exception $th) {
                        // do nothing
                    }
                }
            }
        }
    }

    /**
     * prepend new PathRepository if url found in correspondance
     * @param RepositoryManager $repositoryManager
     * @param array $correspondances
     * @param VcsRepository $repo
     * @throws Exception if not found
     */
    protected function prependRepoIfFound(
        RepositoryManager $repositoryManager,
        array $correspondances,
        VcsRepository $repo
    )
    {
        $url = $this->getVCSUrl($repo);
        list(
            'projectName' => $projectName,
            'packageName'=>$packageName,
            'path'=>$path,
            'version' =>$version
        ) = $this->findUrlInCorrespondances($correspondances,$url);
        $repositoryManager->prependRepository(new PathRepository(
            [
                'type' => 'path',
                'url' => $path,
                'options' => [
                    'name' => $packageName,
                    'versions' => [
                        $packageName => $version,
                        'symlink' => true
                    ]
                ]
            ],
            $this->io,
            $this->composer->getConfig()
        ));

        $this->io->write("Using symlink \"$path\" to replace VCS $url");
    }

    /**
     * extract available folder from main-config.json
     * @param array $config
     * @return array [$projectName => ['packageName' => string, 'path' => string, 'version' => string ],...] 
     */
    protected function extractAvailableCorrespondances(array $config): array
    {
        $correspondances = [];
        if (isset($config['local-repositories']) && is_array($config['local-repositories'])){
            foreach($config['local-repositories'] as $name => $packageName){
                $version = 'dev-master';
                if (is_array($packageName)){
                    $version = (!empty($packageName['version']) && is_string($packageName['version'])) ? $packageName['version'] : $version;
                    $packageName = $packageName['packageName'] ?? '';
                }
                if (is_string($name) 
                        && !empty(basename($name)) 
                        && is_string($packageName) 
                        && !empty($packageName) 
                        && !in_array(basename($name),['.','..'])){
                    $projectName = basename($name);
                    $dirName = "../$projectName";
                    if (is_dir($dirName)){
                        $correspondances[$projectName] = [
                            'packageName' => $packageName,
                            'path' => $dirName,
                            'version' => $version
                        ];
                    }
                }
            }
        }
        return $correspondances;
    }

    /**
     * getUrlFromVcsRepo
     * @param VcsRepository $repo
     * @return string
     * @throws Exception
     */
    protected function getVCSUrl(VcsRepository $repo): string
    {
        $repoName = $repo->getRepoName();
        $start = 'vcs repo (';
        $end = ')';
        if (substr($repoName,0,strlen($start)) != $start || substr($repoName,-strlen($end)) != $end){
            throw new Exception("VCS' repoName format not recognized !");
        }
        $url = substr($repoName,strlen($start),-strlen($end));
        if (strpos($url,' ') === false){
            throw new Exception("VCS' repoName format not recognized !");
        }
        list($driverType,$url) = explode(' ',$url,2);
        return $url;
    }

    /**
     * search url in corespondances
     * @param array $correspondances
     * @param string $url
     * @return array
     * @throws Exception
     */
    protected function findUrlInCorrespondances(array $correspondances, string $url): array
    {
        foreach ($correspondances as $projectName => $data) {
            if (substr($url,-strlen($projectName)-1) == "/$projectName"){
                return array_merge([
                    'projectName' => $projectName
                ],$data);
            }
        }
        throw new Exception("Correspondance not found");
    }

    /* === implements Capable === */
    public function getCapabilities()
    {
        return [
            'Composer\Plugin\Capability\CommandProvider' => 'Seacms\\Command\\SeacmsTest',
        ];
    }
    /* === implements EventsSubscriberInterface === */
    public static function getSubscribedEvents()
    {
        return [
            ScriptEvents::PRE_UPDATE_CMD => 'onPreUpdate'
        ];
    }
}